//	bfe2 - control buttons
//	Copyright (c) 1999-2003 Brand Huntsman and Lee Salzman
//


#include "common.h"
#include "functions.h"


//////////////////////////////////////////////////////////////////////////

// global
GtkWidget *address_text, *offset_text, *instruction_text;
uint32 i_virtual_segment, i_virtual_offset, i_physical_address;
char i_bytes[LEN_INSTRUCTION], i_description[LEN_INSTRUCTION];

// local
GtkWidget *go_button, *stop_button, *step_button, *next_button;
uint i_size, enable_next, next_active;
uint32 next_physical;

//////////////////////////////////////////////////////////////////////////


void string_read( char *buffer, char *string, const char delimiter ){
	register char *current, *end, c;
	register uint i;

	i = 0;
	current = string;
	end = string + LEN_INSTRUCTION;

	while(current < end){
		c = buffer[i];
		i++;
		if((c == delimiter) || (c == EOF)){
			do{
				--current;
			} while((current >= string) && (isspace(*current) != 0));
			*(++current) = '\0';
			return;
		}
		*current = c;
		++current;
	}
	*(--end) = '\0';
	return;
}


void set_ctrl_running( gboolean v ){
	if(v == TRUE){
		gtk_entry_set_text(GTK_ENTRY(address_text), "");
		gtk_entry_set_text(GTK_ENTRY(offset_text), "");
		gtk_entry_set_text(GTK_ENTRY(instruction_text), ".. r u n n i n g ..");

		gtk_widget_set_sensitive(stop_button, TRUE);
		v = FALSE;
	} else {
		gtk_widget_set_sensitive(stop_button, FALSE);
		v = TRUE;
	}

	gtk_widget_set_sensitive(go_button, v);
	gtk_widget_set_sensitive(step_button, v);
	if(enable_next) gtk_widget_set_sensitive(next_button, v);
	else gtk_widget_set_sensitive(next_button, FALSE);
	gtk_widget_set_sensitive(refreshall_button, v);

	breakpointsEnablePage(v);
	watchpointsEnablePage(v);
	memoryEnablePage(v);
	structuresEnablePage(v);
	prefsEnablePage(v);
}


void ctrl_go( GtkWidget *widget, gpointer data ){
	bochs_running = 1;

	set_ctrl_running(TRUE);

	fprintf(writepipe, "c\n");
	fgetc(readpipe);
	hook_sigio();

	historyWrite("GO...");
}


void ctrl_stop( GtkWidget *widget, gpointer data ){
	ctrlStop(0);
}


void ctrl_step( GtkWidget *widget, gpointer data ){
	fprintf(writepipe, "s\n");
	ctrlGetInstruction();
}


void ctrl_next( GtkWidget *widget, gpointer data ){
	if(virtual_to_physical(&cpu.cs, i_virtual_offset + i_size, &next_physical)){
		fprintf(writepipe, "vb 0x%X:0x%X\n", i_virtual_segment, i_virtual_offset + i_size);
		#ifdef __DEBUG__
		 g_print("BFE DEBUG: (NEXT) vb 0x%X:0x%X\n", i_virtual_segment, i_virtual_offset + i_size);
		#endif
		prompt_read();

		next_active = 1;
		bochs_running = 1;

		set_ctrl_running(TRUE);

		fprintf(writepipe, "c\n");

		hook_sigio();

		historyWrite("NEXT...");
	} else historyWrite("ERROR: can't enable NEXT");
}


uint read_instruction( ){
	char buffer[LEN_BUFFER], address[LEN_ADDRESS], *word;
	uint i;

	enable_next = 0;
	gtk_widget_set_sensitive(next_button, FALSE);

	if(bochs_offline) return(1);

	i = 1;
	fgets(buffer, LEN_BUFFER, readpipe); // get "Next at t=1"
	fgets(buffer, LEN_BUFFER, readpipe); // get instruction

	// get instruction address
	if(sscanf(&buffer[i], "%*s %x:%x: ", &i_virtual_segment, &i_virtual_offset) != 2) return(1);
	goto_next_word(buffer, &i);
	goto_next_word(buffer, &i);
	snprintf(address, LEN_ADDRESS, "%.4X : %.8X", i_virtual_segment, i_virtual_offset);
	gtk_entry_set_text(GTK_ENTRY(offset_text), address);

	// get instruction size
	if(buffer[i] == '('){
		goto_next_word(buffer, &i);
		goto_next_word(buffer, &i);
	}
	sscanf(&buffer[i], "%*x%n: ", &i_size);
	i_size /= 2;
	string_read(&buffer[i], i_bytes, ':');
	goto_next_word(buffer, &i);

	// get instruction
	string_read(&buffer[i], i_description, '\n');
	if(i_description == NULL) return(1);
	gtk_entry_set_text(GTK_ENTRY(instruction_text), i_description);

	// enable NEXT button for special instructions
	get_word(buffer, &i, &word);
	if(string_compare(word, "call") || string_compare(word, "int")
	   || string_compare(word, "rep:") || string_compare(word, "loop")
	   || (word[0] == 'j' && !string_compare(word, "jmp"))){
		enable_next = 1;
		gtk_widget_set_sensitive(next_button, TRUE);
	}

	return(0);
}


//////////////////////////////////////////////////////////////////////////


void ctrlInit( GtkWidget *box ){
	GtkWidget *hbox;

	hbox = new_hbox(box, FALSE);

	// address text field
	address_text = new_text_entry(hbox, FALSE, WLEN_REGISTER);
	gtk_entry_set_editable(GTK_ENTRY(address_text), FALSE);

	// offset text field
	offset_text = new_text_entry(hbox, FALSE, WLEN_DTABLE);
	gtk_entry_set_editable(GTK_ENTRY(offset_text), FALSE);

	// instruction text field
	instruction_text = new_text_entry(hbox, TRUE, 0);
	gtk_entry_set_editable(GTK_ENTRY(instruction_text), FALSE);

	// go button
	go_button = new_button(hbox, FALSE, " Go ");
	gtk_signal_connect(GTK_OBJECT(go_button), "clicked", GTK_SIGNAL_FUNC(ctrl_go), NULL);

	// stop button
	stop_button = new_button(hbox, FALSE, " Stop ");
	gtk_signal_connect(GTK_OBJECT(stop_button), "clicked", GTK_SIGNAL_FUNC(ctrl_stop), NULL);

	// step button
	step_button = new_button(hbox, FALSE, " Step ");
	gtk_signal_connect(GTK_OBJECT(step_button), "clicked", GTK_SIGNAL_FUNC(ctrl_step), NULL);

	// next button
	next_button = new_button(hbox, FALSE, " Next ");
	gtk_signal_connect(GTK_OBJECT(next_button), "clicked", GTK_SIGNAL_FUNC(ctrl_next), NULL);

	enable_next = 0;
	next_active = 0;
}


void ctrlStop( uint sigio ){
	char buffer[LEN_BUFFER], watch[10];
	uint bp, n, i;

	unhook_sigio();
	bochs_running = 0;

	if(!sigio){
		// stop button was pressed, send ctrl-C to bochs
		kill(bochs_pid, SIGINT);
		ctrlGetInstruction();
	} else {
		// read breakpoint info
		fgets(buffer, LEN_BUFFER, readpipe);
		if(sscanf(buffer, "%*s Breakpoint %u,", &bp) == 1){
			ctrlGetInstruction();
			if(next_active && i_physical_address == next_physical){
				// delete breakpoint that was used by the NEXT button
				fprintf(writepipe, "d %u\n", bp);
				prompt_read();
				next_active = 0;
			} else {
				// encountered a normal breakpoint, mark in list
				historyWrite("Stopped on a breakpoint");
				breakpointsUpdate(BP_REDRAW);
				for(n = 0; n < MAX_BREAKPOINTS && breakpoints[n].enabled; n++)
					if(bp == breakpoints[n].number) break;
				breakpointsSelect(n);
			}
		} else if(sscanf(buffer, "%*s Caught %s watch point", watch) == 1){
			ctrlGetInstruction();

  // TODO: bochs doesnt return any info on the watchpoint so bfe has no way to find and mark it in the list

			// encountered a watchpoint, mark in list
			historyWrite("Stopped on a watchpoint");
			watchpointsUpdate(BP_REDRAW);
//			if(string_compare(watch, "read")) wp = WP_READ;
//			else wp = WP_WRITE;
//			for(n = 0; n < MAX_WATCHPOINTS && watchpoints[n].enabled; n++)
//				if(wp == watchpoints[n].type && ??? == watchpoints[n].offset) break;
//			watchpointsSelect(n);

			// hack to mark watchpoint if only one is set
			for(n = 0; n < MAX_WATCHPOINTS && watchpoints[n].enabled; n++);
			if(n == 1) watchpointsSelect(0);
		} else if(tracing){
			// tracing is on
			i = 1;
			sscanf(&buffer[i], "%*s %x:%x: ", &i_virtual_segment, &i_virtual_offset);
			goto_next_word(buffer, &i);
			goto_next_word(buffer, &i);
			if(buffer[i] == '('){
				goto_next_word(buffer, &i);
				goto_next_word(buffer, &i);
			}
			sscanf(&buffer[i], "%*x%n: ", &i_size);
			i_size /= 2;
			string_read(&buffer[i], i_bytes, ':');
			goto_next_word(buffer, &i);
			string_read(&buffer[i], i_description, '\n');

  // BUG

			//stateUpdate();
			//if(!virtual_to_physical(&cpu.cs, i_virtual_offset, &i_physical_address))
			//	i_physical_address = 0;

			historyUpdate();
			bochs_running = 1;
			hook_sigio();
			return;
		} else {
			// dont know what happened here
		}
	}

	set_ctrl_running(FALSE);
}


void ctrlEnableControls( gboolean v ){
	if(v == FALSE){
		gtk_entry_set_text(GTK_ENTRY(address_text), "");
		gtk_entry_set_text(GTK_ENTRY(offset_text), "");
		gtk_entry_set_text(GTK_ENTRY(instruction_text), ".. o f f l i n e ..");

                gtk_widget_set_sensitive(stop_button, FALSE);
	}

	gtk_widget_set_sensitive(go_button, v);
	gtk_widget_set_sensitive(step_button, v);
	if(enable_next) gtk_widget_set_sensitive(next_button, v);
	else gtk_widget_set_sensitive(next_button, FALSE);
}


void ctrlGetInstruction( ){
	char address[LEN_ADDRESS];

	if(bochs_offline) return;

	#ifdef __DEBUG__
	 g_print("BFE DEBUG: (UPDATE) prompt = %u\n", p_step);
	#endif

	read_instruction();
	prompt_read();

	stateUpdate();

	if(!virtual_to_physical(&cpu.cs, i_virtual_offset, &i_physical_address))
		i_physical_address = 0;
	snprintf(address, LEN_ADDRESS, "%.8X", i_physical_address);
	gtk_entry_set_text(GTK_ENTRY(address_text), address);

	historyUpdate();
}
